/* Merge and MergeSort all in one file */
/* This is the general version of (iterative) mergesort which 
 *  works regardless of whether the input is a power of 2. 
 * This is written to allow the user to input a number (length
 * of desired array); then the program allocates space for an
 * array of that length and fills with random numbers, then
 * it calls mergesort.  It outputs statistics on running time  */

/* The 'main' code below looks a tiny bit different than the 
 * code presented in slide 8, lecture 20.  That's because 
 * to save space (and to focus on 'calloc' and the measurement
 * of running time) I removed the 'if'-'else' wrapper which 
 * printed out the details in the case of small arrays. */

/* For answering the 2nd question of the week-8 tutorial, we
 * must edit the 'main'.  Remove the 'if'-'else' wrapper and 
 * the code for sz < 20. 
 * Also get rid of the first-printf and the scanf, because we
 * will not take input from the user - we will do a bunch of 
 * iterations with different values for sz.
 * Then you should put in a for-loop around the code, starting
 * just before 'calloc', and ending just after the 'free'.  
 * The 'for' should start with a fairly big value of 'sz' 
 * (maybe 10000 or 20000), and after each iteration, should
 * double this value.  The number of doublings possible
 * will depend on space available for creating the array. 
 * Use trial and error to see how many times is possible.
 * It is actually possible to declare very large arrays 
 * (on DICE anyway) */


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>  /* For using the 'assert' statement. */ 
#include <time.h>    /* for the measuring of running-time */

/* Merge function - note use of k++, i++ sitting *within* other
 * expressions.  This is a very delicate use of k++ (big contrast
 * to if we had used ++k).  Please make sure you are ok with
 * the details of these (within expression) increments.  */  
void merge(int a[], int b[], int c[], int m, int n) {
  int i=0, j=0, k=0;
  while (i < m && j < n) {
    if (a[i] <= b[j])
      c[k++] = a[i++];
    else 
      c[k++] = b[j++];
  }
  /* At this stage, done comparing.  All the rest of c will come
   * from either a, or else b, depending on which has stuff left. */
  while (i < m) 
    c[k++] = a[i++];
  while (j < n)
    c[k++] = b[j++];
}

void mergesort(int key[], int n){
  int j, k, *w;
  w = calloc(n, sizeof(int));       /* space for temporary array */
  assert (w != NULL);
  if ((n % 2) == 1)
    w[n-1] = key[n-1]; 
  for (k = 1; k < n; k *= 2) {
    for (j = 0; j < n - 2*k; j += 2*k) 
      /* Do merging as usual for all j < n - 2*k. 
       * Note that we stop sooner than in power-of-2 case, to 
       * make sure call merge correctly at the (possibly-awkwardly
       * length-ed) end-of-array.    */
      merge(key + j, key + j + k, w + j, k, k);
    if (n-j > k) 
      /* Case where more-work-needed. Sizes different: k, n-j-k. */
      merge(key + j, key + j + k, w + j, k, (n-j)-k);
    for (j = 0; j < n; ++j)
      key[j] = w[j];
    /* Commenting out the piece of code which outputs the 
     * initial, and then sorted, array.  Confusing if running
     * running-time-tests on large arrays.  
     * Uncomment it if you decided you want to see the iterative 
     * progress of mergesort on smallish arrays.
     *  for (j = 0; j < n; ++j) 
     *  printf("%4d%s", key[j], ((j < n -1) ? "" : "\n"));  */
  }
  free(w);        /* Free up the dynamic memory no longer in use. */
}

/* Function to write-out the contents of key[] in appropriate form. */ 
void wrt(int key[], int sz) {
  int i;
  for (i = 0; i < sz; ++i) 
    printf("%4d%s", key[i], ((i < sz -1) ? "" : "\n"));
}


int main(void) {
  int i, sz, *key;
  double start, stop, t;        
  printf("Input desired size of array: "); 
  scanf("%d", &sz);
  printf("\n");
  key = calloc(sz, sizeof(int));  /* Make array of this size */
  if (key != NULL) {              /* check there was space */
    for(i = 0; i < sz; i++)      /* Fill array: */ 
      key[i] = rand() % 1000;    /* rand() returns 1 random int */
    if(sz < 20) {
      /* For small arrays, show before-and-after array on the screen */
      printf("Before mergesort: \n");
      wrt(key, sz);
      mergesort(key, sz);
      printf("After mergesort: \n");
      wrt(key, sz);
    }
    else {
      /* For large arrays, our interest is in the runtime statistics. */ 
      start = (double)clock();
      mergesort(key, sz);
      stop = (double)clock();
      t = (stop-start)/CLOCKS_PER_SEC;
      printf("Time on array of length %d was %f sec.\n", sz, t);
    }
    free(key);
  }
  else printf("Trouble allocating space for that many integers.\n");
  return EXIT_SUCCESS;
}
